require("prototypes.recycler")

local floor = math.floor
local max = math.max
local sqrt = math.sqrt
local huge = math.huge
local next = next
local pairs = pairs


local acceptable_recipe_categories = {
	["crafting"]=true,
	["advanced-crafting"]=true,
	["smelting"]=true,
	["packing"]=true,
	["alien-crafting"]=true,
	["crafting-with-fluid"]=true,
	}

local whitelist={
	["battery"]=true,
}


local function findfirst(element,list)
	for k,v in pairs(list) do
		if v==element then return k end
	end
	return false
end


local lootentry
for _, v in pairs(data.raw["tree"]) do
	if v.minable and v.minable.result == "raw-wood" then
		local result=floor((v.minable.count+1)/2)
		if result>0 then
			local lootentry = {count_max = result+.4, count_min = 1, item = "raw-wood", probability = 1/(1+result)}
			if v.loot == nil then
				v["loot"] = {lootentry}
			else
				table.insert(v.loot,lootentry)
			end
		end
	end
end
	
local nillist={}

local max_modules={}
for _,category in pairs(data.raw["recipe-category"]) do
	local m=0
	for _,machine in pairs(data.raw["assembling-machine"]) do
		if (machine.module_slots or 0)>m then m=machine.module_slots end
	end
	for _,machine in pairs(data.raw["furnace"]) do
		if (machine.module_slots or 0)>m then m=machine.module_slots end
	end
	max_modules[category.name]=m
end

local function MaxProductivity(rname)
	local maxmod = 0
	for _,module in pairs(data.raw.module) do
		if module.effect and module.effect.productivity and (not module.limitation or findfirst(rname,module.limitation)) and module.effect.productivity.bonus>maxmod then
			maxmod = module.effect.productivity.bonus
		end
	end
	return 1+(maxmod*(max_modules[data.raw.recipe[rname].category or "crafting"] or 0))
end

local recipe_cache={}

local itemlevel={} --considered a base resource, even if recipes yield it
for _,resource in pairs(data.raw.resource) do
	if resource.minable and resource.minable.result then
		recipe_cache[resource.minable.result]=nillist
		itemlevel[resource.minable.result]=0
	end
end

--[[
    ingredients =
    {
      {type="fluid", name="crude-oil", amount=25},
      {type="item", name="empty-barrel", amount=1},
    },
    ingredients =
    {
      {"iron-gear-wheel", 3},
      {"stone-furnace", 1},
      {"iron-plate", 3}
    },
--]]

--ultimate yield: itemrecipe as clean tree itemrecipe[item-name] = {[ingredient]=rate}
--[[

run through all recipes
	if recipe has one item result, add to
		result_of[recipename]=result
		ingredients_for[recipename]={ingredient1=scaledcount, ingredient2=scaledcount}
			recipes_using[ingredientname]={recipe,recipe}
		add recipename to recipes_for[result]
		
	
run through all items
	if recipesfor[item]=nil, set itemlevel[item]=0
	if not itemlevel[item] then
		needslevel[item]=true


--]]


local needed = {}	--list of all items that will need to be (but aren't in) in recipe_cache
local result_of, ingredients_for = {},{} --intermediate
local recipes_using, recipes_for = {}, {} --trackers

local this_result,this_count,ingredientlist, this_thing

for _,recipe in pairs(data.raw.recipe) do
	if not recipe.category or acceptable_recipe_categories[recipe.category] or whitelist[recipe.name] then
		this_result=recipe.result
		if this_result then
			this_count = recipe.result_count or 1
		else
			for _,result in pairs(recipe.results) do
				if result.type=="item" then
					if this_result then
						this_result = nil
						break
					else
						this_result = result.name or result[1]
						this_count = result.amount or result[2] or 1
					end
				end
			end
		end
		if this_result then --we have an overridden recipe with one result
			this_count = this_count * MaxProductivity(recipe.name)
			ingredientlist={}--[name]=totalamount
			for _,ingredient in pairs(recipe.ingredients) do
				if ingredient.type then
					if ingredient.type=="item" then
						ingredientlist[ingredient.name] = (ingredientlist[ingredient.name] or 0) + (ingredient.amount or 1)/this_count
					end
				else
					ingredientlist[ingredient.name or ingredient[1]] = (ingredientlist[ingredient.name or ingredient[1]] or 0) + ((ingredient.amount or ingredient[2]) or 1)/this_count
				end
			end
			if next(ingredientlist) ~= nil then --valid input(s)
				this_thing=recipe.name
				result_of[this_thing] = this_result
				ingredients_for[this_thing] = ingredientlist
				for i,c in pairs(ingredientlist) do
					if recipes_using[i] then
						table.insert(recipes_using[i],this_thing)
					else
						recipes_using[i]={this_thing}
					end
				end
				if recipes_for[this_result] then
					table.insert(recipes_for[this_result],this_thing)
				else
					recipes_for[this_result] = {this_thing}
				end
			end
		end
	end
end


--filled result_of, ingredients_for = {},{} --intermediate
--  	 recipes_using, recipes_for = {}, {} --trackers

--itemlevel[itemname]=0
--recipe_cache[item]=nillist for bases

for _,i in pairs(data.raw.item) do
	if recipes_for[i.name] and #recipes_for[i.name] > 0 then
		needed[i.name]=true
	else
		if not itemlevel[i.name] then itemlevel[i.name]=0 end
		needed[i.name]=nil
	end
end

this_thing = {}
for item,level in pairs(itemlevel) do
	if level<1 then
		this_thing[item]=true
		needed[item]=nil
	end --should all be 0 at this point but let's just make sure
end

--error(serpent.block(result_of))
--error(serpent.block(ingredients_for))
--error(serpent.block(recipes_using))
--error(serpent.block(recipes_for))

--local logger,itercount={},0


local levelcum,lowest,ref = {}
local next_thing={}

while next(needed)~=nil do

--[[
	itercount = itercount + 1
	table.insert(logger,"-------this_thing-----"..itercount)
	table.insert(logger,serpent.block(this_thing))
	table.insert(logger,"------levels------")
	table.insert(logger,serpent.block(itemlevel))
--]]
	if next(this_thing)==nil then
		error("Scrap Mod error: \"recipe parse incomplete,\" probably an incompatibility with another mod. Please let me know in the thread at http://www.factorioforums.com/forum/viewtopic.php?f=14&t=3612 or contact Schmendrick on the Factorio forums.")
		--table.insert(logger,"-=-=-=-=-=-=-=-=-=preparsed recipes=-=-=-=-=-=-=-=-=-=-=-=-"..serpent.block(ingredients_for))
		--table.insert(logger,"---STILL NEEDED--\n"..serpent.block(needed))
		--table.insert(logger,"==================================final cache==========================="..serpent.block(recipe_cache))
		--error(table.concat(logger,"\n"))
	end
	

	for baseitem in pairs(this_thing) do
		if recipes_using[baseitem] then
			for _,recipe in pairs(recipes_using[baseitem]) do if itemlevel[result_of[recipe]]==nil then
				ref={0,0}
				for item,q in pairs(ingredients_for[recipe]) do
					if ref and itemlevel[item] then
						if itemlevel[item] >= ref[1] then
							ref[1]=itemlevel[item]+1
						end
						ref[2]=ref[2]+q*(itemlevel[item]+1)
					else
						ref = nil
					end
				end
				if ref and (levelcum[recipe]==nil or levelcum[recipe][1]>ref[1] or levelcum[recipe][1]==ref[1] and levelcum[recipe][2]>ref[2]) then
					levelcum[recipe]=ref
					--table.insert(logger,recipe.." tier "..serpent.line(ref))
				end
			end end
			--just finished establishing all (new possible) levelcum's for recipes using baseitem
			
			for _,recipe in pairs(recipes_using[baseitem]) do if itemlevel[result_of[recipe]]==nil then
				lowest,ref={huge,huge},nil
				for _,recipe in pairs(recipes_for[result_of[recipe]]) do
					local level = levelcum[recipe]
					if level and (level[1]<lowest[1] or (level[1]==lowest[1] and level[2]<lowest[2])) then
						lowest=level
						ref=recipe
					end
				end
				
				if ref then
					--table.insert(logger,"added recipe "..ref..serpent.line(ingredients_for[ref]))
					recipe_cache[result_of[ref]]=ingredients_for[ref]
					itemlevel[result_of[ref]]=lowest[1]
					--table.insert(logger,"set itemlevel of "..result_of[ref].." to "..lowest[1])
					needed[result_of[ref]]=nil
					next_thing[result_of[ref]]=true
				end
			end end
		end
	end
	this_thing=next_thing
	next_thing={}
--error(serpent.block(this_thing))
end
--error(serpent.block(levelcum))
--error(serpent.block(recipe_cache).."\nitemlevels "..serpent.block(itemlevel))
--error(serpent.block(recipes_using))

--------------------FINALLY RECIPES PARSED----
local recycle_recipes={}

local function addtolist(a,b,rate)
	for k,v in pairs(b) do
		a[k]=(a[k] or 0)+v*rate
	end
	
	return a
end

local function get_raw(item)
	if recycle_recipes[item] then return recycle_recipes[item] end
	if itemlevel[item]<1 then return {[item]=1} end
	local result={}
	if not recipe_cache[item] then error("get_raw error "..item) end
	for ing,q in pairs(recipe_cache[item]) do
		if itemlevel[ing]<1 then
			result[ing]=(result[ing] or 0)+q
		else
			addtolist(result,get_raw(ing),q)
		end
	end
	recycle_recipes[item]=result
	return result
end

for item in pairs(itemlevel) do
	if recycle_recipes[item]==nil then get_raw(item) end
end

--error(table.concat(logger,"\n"))
--error(serpent.block(recipe_cache))
--logger={}
--table.insert(logger,serpent.block(recycle_recipes))
--error(serpent.block(recycle_recipes))

 --logger={}

local scrapmax = data.raw.item["metal-scrap"].stack_size
local copper,iron,scrap
local entry
for item,scraplist in pairs(recycle_recipes) do if itemlevel[item]>1 then
	copper=scraplist["copper-ore"] or 0
	iron=scraplist["iron-ore"] or 0
	if (copper+iron)>0 then
		
entry={ type = "recipe",
			--name = "melt-pistol",
			--icon = "__plowshares__/graphics/scrap.png",
			category = "recycling",
			hidden = "true",
			enabled = "true",
			--energy_required = 5,
			--ingredients = {{ "pistol", 1}},
			--result = "metal-scrap",
			--result_count = 4
		  }
		entry.name="melt-"..item
		if copper<iron then scrap=copper else scrap=iron end		
		if scrap>=data.raw.item["metal-scrap"].stack_size/2 or (scrap*2>=iron and scrap*2>=copper) then
			entry.result="metal-scrap"
		else
			if iron>copper then
				entry.result="iron-ore"
				scrap=iron
			else
				entry.result="copper-ore"
				scrap=copper
			end
		end
		entry.icon=data.raw.item[entry.result].icon
		scrap=scrap*.9--semiarbitrary recycle loss cost (cap)
		--still need energy_required, ingredient(count):result_count
		--find good numerator and denominator for scrap. use copper=topper(numerator)=output and iron=denominator=input
		local subgroupchecks={"item","ammo","armor","repair-tool","capsule","gun"}
		local maxin,maxout=0,0
		for _,subgroup in pairs(subgroupchecks) do
			if data.raw[subgroup][item] then maxin=data.raw[subgroup][item].stack_size end
		end
		
		for _,subgroup in pairs(subgroupchecks) do
			if data.raw[subgroup][entry.result] then maxout=data.raw[subgroup][entry.result].stack_size end
		end
		
		local erof
		
		if data.raw.ammo[item] or data.raw["repair-tool"][item] then
			maxin = maxin - 1
		end
		
		if scrap<0 then
			iron = floor(1/scrap+1.9)-1
		else
			iron = 0
		end
		
		repeat
			iron = iron + 1
			copper,erof = math.modf(iron * scrap)
		until (erof/copper)<.1 or iron >= maxin or copper >= maxout
		
		if data.raw.ammo[item] or data.raw["repair-tool"][item] then
			iron = iron + 1
		end
		
		
		if copper>0 then
			if copper>maxout then copper=maxout end
			
			entry.ingredients={{item,iron}}
			entry.result_count=copper
			entry.energy_required=floor(sqrt(copper*10))/2
			
			--[[
			entry.result=nil
			entry.result_count=nil
			
			entry.results={
			{type="item",name="copper-ore",amount=1},
			{type="item",name="iron-ore",amount=1},
			}
			--]]

			
			data:extend({entry})
			--table.insert(logger,table.concat({item,iron,entry.result,copper,entry.energy_required},","))
		end
	end
end end

--error(table.concat(logger,"\n"))

--[[
  { type="recipe-category", name= "alien-crafting" },
  { type="recipe-category", name= "forge" },
  { type="recipe-category", name= "blast-furnace" },
  { type="recipe-category", name= "pressure-vessel" },
  { type="recipe-category", name= "electrolycer" },
  { type="recipe-category", name= "drill" },
  { type="recipe-category", name= "drilling-tower" },
  { type="recipe-category", name= "robot" },]]
 --[[WARNING: if new recipe types added, update]]


local totalmax = 0

local function add_ingredients(base,add,rate)
	rate = rate or 1
	if rate < .001 then return base end
	local need, safeadd
	for _,additem in pairs(add) do
		if additem.type ~="fluid" --[[WARNING: if new ingredient types added, update]] then
			safeadd={additem.name or additem[1], additem.amount or additem[2]}
			need = true
			for _,matchitem in pairs(base) do				
				if matchitem[1] == safeadd[1] then
					need = false
					matchitem[2] = matchitem[2] + safeadd[2] * rate
					break
				end
			end
			if need then table.insert(base, {safeadd[1],safeadd[2]*rate} ) end
		end
	end
	return base
end

-- --[[
local minables = {}
local entity_name, ingredient_queue, lootlist, base_ingredients, need, placeable

----
for _,category in pairs(data.raw) do
	for _,candidate in pairs(category) do
		if candidate.minable then
			minables[candidate.name] = candidate
		end
	end
end

local function mostitems(item,reserve,num)
	local thismost
	if not recycle_recipes[item] then
		return reserve[ing] or 0
	end
	for ing,q in pairs(recycle_recipes[item]) do
		if not reserve[ing] then return 0 end
		thismost = reserve[ing]/q
		if thismost<num then num=thismost end
	end
	return num
end

local softcap, hardcap
local thisitem, adding
local deadlist

--logger={}
-------------------------------------------------
for result,ingredients in pairs(recipe_cache) do
	if data.raw.item[result] then
	entity_name = data.raw.item[result].place_result
	placeable = minables[entity_name]
	if placeable then
		dimret=itemlevel[result]
		--1/2+ + 1/sqrt(dimret) level=25; 1/5
		if dimret==0 then dimret=1 else dimret=.5+1/sqrt(4*dimret) end --initial devalue of first tier
		totalmax=0
		softcap = 7 * (placeable.selection_box[1][1] - placeable.selection_box[2][1]) * (placeable.selection_box[1][2] - placeable.selection_box[2][2])
		if softcap<0 then softcap=-softcap end
		
		lootlist={}
		deadlist = {}
		ingredient_queue=ingredients
		hardcap = {}
		for i,q in pairs(recycle_recipes[result]) do
			if q > 50 then hardcap[i]=50 else hardcap[i]=q end
		end
		
		while next(ingredient_queue)~=nil do
			next_queue={}
			
			for ing,num in pairs(ingredient_queue) do
				thisitem=mostitems(ing,hardcap,num)
				adding=thisitem*dimret
				thisitem=thisitem-adding
				if adding>0 then
					lootlist[ing] = (lootlist[ing] or 0)+adding
					for i,q in pairs(recycle_recipes[ing]) do
						hardcap[i]=hardcap[i]-q*adding
					end
					local okay=true
					for _,q in pairs(recipe_cache[ing]) do
						if q*thisitem < 0 then okay=false end
					end
					if okay then
						for i,q in pairs(recipe_cache[ing]) do
							next_queue[i]=(next_queue[i] or 0)+q*thisitem
						end
					else
						
					end
					--addtolist(next_queue,recipe_cache[ing],thisitem)
					--next_queue[ing]=(next_queue[ing] or 0) + adding
					totalmax=totalmax+adding
				end
				
			end
		
			ingredient_queue=next_queue
			dimret = softcap/(softcap+totalmax)
		end

		--flush
		for n=itemlevel[result]-1,1,-1 do
			for i,q in pairs(lootlist) do if itemlevel[i]==n then
				if q<1 then
					if itemlevel[i]<1 then
						for si,sq in pairs(recipe_cache[i]) do
							lootlist[si]=(lootlist[si] or 0)+si*q
						end
					end
					--lootlist[q]
				end
			end end
		end
		--mark nils
		for i,q in pairs(lootlist) do
			if q<1 then deadlist[i]=true end
		end
		
		
		--regenerate
		dimret=itemlevel[result]
		--1/2+ + 1/sqrt(dimret) level=25; 1/5
		if dimret==0 then dimret=1 else dimret=.5+1/sqrt(4*dimret) end --initial devalue of first tier
		totalmax=0
		softcap = 7 * (placeable.selection_box[1][1] - placeable.selection_box[2][1]) * (placeable.selection_box[1][2] - placeable.selection_box[2][2])
		if softcap<0 then softcap=-softcap end
		
		lootlist={}
		ingredient_queue=ingredients
		hardcap = {}
		for i,q in pairs(recycle_recipes[result]) do
			if q > 50 then hardcap[i]=50 else hardcap[i]=q end
		end
		
		while next(ingredient_queue)~=nil do
			next_queue={}
			
			for ing,num in pairs(ingredient_queue) do
			
				if deadlist[ing] then
					lootlist[ing]=(lootlist[ing] or 0)+num--*dimret
				else
			
					thisitem=mostitems(ing,hardcap,num)
					adding=thisitem*dimret
					thisitem=thisitem-adding
					if adding>0 then
						lootlist[ing] = (lootlist[ing] or 0)+adding
						for i,q in pairs(recycle_recipes[ing]) do
							hardcap[i]=hardcap[i]-q*adding
						end
						local okay=true
						for _,q in pairs(recipe_cache[ing]) do
							if q*thisitem < 0 then okay=false end
						end
						if okay then
							for i,q in pairs(recipe_cache[ing]) do
								next_queue[i]=(next_queue[i] or 0)+q*thisitem
							end
						else
							
						end
						--addtolist(next_queue,recipe_cache[ing],thisitem)
						--next_queue[ing]=(next_queue[ing] or 0) + adding
						totalmax=totalmax+adding
					end
				end
			end
		
			ingredient_queue=next_queue
			dimret = softcap/(softcap+totalmax)
		end
		
		for n=itemlevel[result]-1,1,-1 do
			for i,q in pairs(lootlist) do if itemlevel[i]==n then
				if q<1 then
					if itemlevel[i]<1 then
						for si,sq in pairs(recipe_cache[i]) do
							lootlist[si]=(lootlist[si] or 0)+si*q
						end
					end
					--lootlist[q]
				end
			end end
		end

--[[
      {
        count_max = 10,
        count_min = 2,
        item = "alien-artifact",
        probability = 1
      }

--]]

		
		if next(lootlist)~=nil then
			--table.insert(logger,result..":"..serpent.block(lootlist))
			
			if not placeable.loot then
				placeable.loot={}
			end
			-- 1 2 3:::: .2, 2/3, 1/4
			for i,q in pairs(lootlist) do
				--q=floor(q)
				if q>1 then
					table.insert(placeable.loot,{item=i,count_min=1,count_max=floor(q),probability=q/(q+1)})
				end
			end
			--table.insert(logger,placeable.name..serpent.block(placeable.loot))
			--error(serpent.block(placeable))
		end
	end end
end
--error(table.concat(logger,"\n"))
